import tornado.escape
import tornado.ioloop
import tornado.web
from . import worker
from . import util
import threading
from abc import ABCMeta
import logging
import asyncio
import numpy as np
import pandas as pd
import tornado.httpserver
import tornado.netutil
import tornado.websocket
from tornado_swagger.components import components
from tornado_swagger.setup import setup_swagger
import re
import json

l = logging.getLogger("pandem-api")


class apiREST(worker.Worker):
    __metaclass__ = ABCMeta 

    def __init__ (self, name, orchestrator_ref, settings):
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)
        self._api_port = settings["pandem"]["source"]["api"]["port"]
    
    def on_start(self):
        super().on_start()
        self._storage_proxy = self._orchestrator_proxy.get_actor('storage').get().proxy()
        self._variables_proxy = self._orchestrator_proxy.get_actor('variables').get().proxy()
        self._pipeline_proxy = self._orchestrator_proxy.get_actor('pipeline').get().proxy()
        try:
            l.debug("Starting HTTP server")
            sockets = tornado.netutil.bind_sockets(self._api_port) 
            self.server = HttpServer(
                storage_proxy=self._storage_proxy,
                variables_proxy = self._variables_proxy,
                pipeline_proxy = self._pipeline_proxy,
                sockets=sockets
            )
            self.server.start()
        except OSError as err:
            l.debug(f"HTTP server startup failed: {err}")

    def on_stop(self):
        self.server.stop()


class HttpServer(threading.Thread):

    def __init__(self, storage_proxy, variables_proxy, pipeline_proxy, sockets): 
        super().__init__()
        self.storage_proxy = storage_proxy
        self.variables_proxy = variables_proxy
        self.pipeline_proxy = pipeline_proxy
        self.sockets = sockets
        self.app = None
        self.server = None
        self.io_loop = None

    def run(self):
        if asyncio:
            # If asyncio is available, Tornado uses it as its IO loop. Since we
            # start Tornado in a another thread than the main thread, we must
            # explicitly create an asyncio loop for the current thread.
            asyncio.set_event_loop(asyncio.new_event_loop())
        self.app = Application(self.storage_proxy, self.variables_proxy, self.pipeline_proxy)
        self.server = tornado.httpserver.HTTPServer(self.app)
        self.server.add_sockets(self.sockets)
        self.io_loop = tornado.ioloop.IOLoop.current()
        self.io_loop.start()

    def stop(self):
        l.debug("Stopping HTTP server")
        self.io_loop.add_callback(self.io_loop.stop)


class SourcesHandler(tornado.web.RequestHandler):
    def initialize(self, storage_proxy):
        self.storage_proxy = storage_proxy
    def get(self):
        """
        ---
        tags:
          - Sources
        summary: List sources
        description: List all sources in sources pickle file within the database.
        operationId: getSources
        responses:
            '200':
              description: List all sources including id, name, last and next execution, current status and step of last job and number of issues of last
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/SourcesModel'
                application/xml:
                  schema:
                    $ref: '#/components/schemas/SourcesModel'
                text/plain:
                  schema:
                    type: string
        """
        df_sources = self.storage_proxy.read_db('source').get()
        df_jobs = self.storage_proxy.read_db('job').get()
        df_issues = self.storage_proxy.read_db('issue').get()
        tags = {}
        sources = {s['name']:s for s in (df_sources.to_dict('records') if df_sources is not None else [])} 
        jobs = df_jobs.to_dict('records') if df_jobs is not None else []
        
        for j in jobs:
          source = sources[j['source']]
          is_active_or_last_job = j["status"] == "in progress" or len([jj for jj in jobs if j["source"] == jj["source"] and jj["start_on"] > j["start_on"]]) == 0
          dls = j['dls_json']
          tag_name = dls["scope"]["tags"][0] if "tags" in dls["scope"] and len(dls["scope"]["tags"]) > 0 else source["name"]
          if tag_name in tags:
            tag = tags[tag_name]
          else:
            tag = {
              "name":tag_name,
              "files":0,
              "size":0,
              "progress":0.0,
              "issues":0,
              "last_import_start":None,
              "last_import_end":None,
              "next_check":None
            }
            tags[tag_name] = tag
          if source["next_exec"] is not None and (tag["next_check"] is None or tag["next_check"] > str(source["next_exec"])):
            tag["next_check"] = str(source["next_exec"])
          
          if is_active_or_last_job and type(j["source_files"]) == list:
            tag["progress"] = (tag["progress"]*tag["files"] + j['progress']*len(j["source_files"]))/(tag["files"] + len(j["source_files"]))
            tag["files"] = tag["files"] + len(j["source_files"])
            tag["size"] =  tag["size"] + sum(j["file_sizes"])
            tag["issues"] =  tag["issues"] + len(df_issues[df_issues['job_id']==j["id"]] if df_issues is not None else [])
            
            if tag["last_import_start"] is None or tag['last_import_start'] < str(j['start_on']):
                tag["last_import_start"] = str(j["start_on"])
            if tag["last_import_end"] is None or (j["end_on"] is not None and str(j["end_on"]) > tag["last_import_end"]):
              tag['last_import_end'] = str(j['end_on'])
            if ("step" not in tag) or tag["progress"] < j["progress"]:
                tag['status'] = j['status']
                tag['step'] = j["step"]
          else:
            if tag["last_import_end"] is None or (j["end_on"] is not None and str(j["end_on"]) > tag["last_import_end"]):
              tag['last_import_end'] = str(j['end_on'])
              
        res = list(tags.values())
        res.sort(key = lambda s:s["name"]) 
        response = {'sources' : res}
        self.write(response)

class JobsBySourceHandler(tornado.web.RequestHandler):
    def initialize(self, storage_proxy, pipeline_proxy):
        self.storage_proxy = storage_proxy
        self.pipeline_proxy = pipeline_proxy
    def get(self):
        """
        ---
        tags:
          - Jobs
        summary: List of jobs related to a source
        description: List of jobs for a particular source including job_id, start, end, status, step, dls file and number of issues.
        operationId: getJobs
        parameters:
          - name: source
            in: query
            description: source name
            required: false
            schema:
              type: string
        responses:
            '200':
              description: List of jobs for a particular source including job_id, start, end, status, step, dls, files, number of issues
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/JobsModel'
                application/xml:
                  schema:
                    $ref: '#/components/schemas/JobsModel'
                text/plain:
                  schema:
                    type: string
        """
        source = self.get_argument('source', default = None)
        df_issues = self.storage_proxy.read_db('issue').get()
        df_jobs = self.storage_proxy.read_db('job').get()
        jobs_list = []
        for j in df_jobs.to_dict('records'):
          dls = j['dls_json']
          tag_name = dls["scope"]["tags"][0] if "tags" in dls["scope"] and len(dls["scope"]["tags"]) > 0 else j["source"]
          if source is None or source == tag_name: 
            jobs_list.append(
              {'id': j['id'],
                'source': str(j['source']),
                'start': str(j['start_on']),
                'end': str(j['end_on']),
                'status': j['status'],
                'step': j['step'],
                #'dls': j['dls_json'],
                'progress': j['progress'],
                'files': len(j['source_files']) if type(j["source_files"]) == list else None,
                'size': sum(j['file_sizes']) if type(j["file_sizes"]) == list else None,
                'issues': len(df_issues[(df_issues['source']==j['source']) & (df_issues['job_id']==j['id'])]) if df_issues is not None else 0
              } 
            )
        jobs_list.sort(key = lambda j:j['id'], reverse = True)      
        response = {'jobs' : jobs_list}
        self.write(response)

    def delete(self):
        """
        ---
        tags:
          - Jobs
        summary: Force a job to be failed and delete all its data
        description: Force a job to be failed and delete all its data
        operationId: failJobs
        parameters:
          - name: job_id
            in: query
            description: job id to delete
            required: true
            schema:
              type: int
        responses:
            '200':
              description: id of deleted job
              content:
                text/plain:
                  schema:
                    type: string
        """
        job_id = self.get_argument('job_id', default = None)
        self.pipeline_proxy.fail_job(int(job_id), delete_job = True).get()
        self.write(job_id)

class IssuesHandler(tornado.web.RequestHandler):
    def initialize(self, storage_proxy):
        self.storage_proxy = storage_proxy
    def get(self):
        """
        ---
        tags:
          - issues
        summary: List of issues realted to a Job of a source
        description: List of issues produced  for a particular job in a source.
        operationId: getIssues
        parameters:
          - name: source
            in: query
            description: source name
            required: false
            schema:
              type: string
          - name: job_id
            in: query
            description: job id
            required: false
            schema:
              type: int
        responses:
            '200':
              description: List of issues found during job of sources
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/IssueModel'
                application/xml:
                  schema:
                    $ref: '#/components/schemas/IssueModel'
                text/plain:
                  schema:
                    type: string
        """
        source = self.get_argument('source', default = None)
        job = self.get_argument('job_id', default = None)

        sources = list(self.storage_proxy.read_db('source').get()['name'])
        if source is not None and source not in sources:
            self.set_status(400)
            return self.write("Invalid or not found source")
        else:
            df_jobs = self.storage_proxy.read_db('job').get()
            if source is not None:
              df_jobs = df_jobs[df_jobs['source']==source]
            if job is not None:
              df_jobs = df_jobs[df_jobs['id'].astype(int)==int(job)]

            df_issues = self.storage_proxy.read_db('issue').get()
            if df_jobs is not None:
              dlss = {j['source']:j['dls_json'] for j in df_jobs.to_dict('records')}
            else :
              dlss = {}
            tags = {source:dls["scope"]["tags"][0] if "tags" in dls["scope"] and len(dls["scope"]["tags"]) > 0 else source for source, dls in dlss.items()}

            if df_issues is not None:
              df_issues = df_issues[df_issues['job_id'].isin(df_jobs['id'])]
              df_issues = df_issues.copy()
              df_issues["raised_on"] = df_issues["raised_on"].astype(str)
              df_issues["tag"] = df_issues['source'].apply(lambda s: tags[s])

              issue_list = df_issues.to_dict('records') 
            else:
              issue_list = []
        response = {'issues' : issue_list}
        self.write(response)


class SourceDetailsHandler(tornado.web.RequestHandler):
    def initialize(self, storage_proxy):
        self.storage_proxy = storage_proxy
    def get(self):
        """
        ---
        tags:
          - Source Details
        summary: List of source details as per data labelling schema
        description: List of sources detauks as oer data labelling schema
        operationId: getSourceDetails
        parameters:
          - name: source
            in: query
            description: source name
            required: false
            schema:
              type: string
        responses:
            '200':
              description: List of data labelling schema for the sources
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/SourceDetailModel'
                application/xml:
                  schema:
                    $ref: '#/components/schemas/SourceDetailModel'
                text/plain:
                  schema:
                    type: string
        """
        source = self.get_argument('source', default = None)
        sources_path = util.pandem_path("files", "source-definitions")
        dls_paths = self.storage_proxy.list_files(sources_path).get()
        dlss = [self.storage_proxy.read_file(f['path']).get() for f in dls_paths if f['path'].endswith(".json")]
        defs = {dls["scope"]["tags"][0]+ " - "+ dls["scope"]["source"] if "tags" in dls["scope"] and len(dls["scope"]["tags"]) > 0 else dls["scope"]["source"]:dls for dls in dlss}
        if source is not None:
          defs = {k:v for k, v in defs.items() if k == source}
        response = {'sources':list(defs.keys()), 'definitions':defs }
        self.write(response)

class VariableListHandler(tornado.web.RequestHandler):
    def initialize(self, storage_proxy, variables_proxy):
        self.storage_proxy = storage_proxy
        self.variables_proxy = variables_proxy
    def get(self):
        """
        ---
        tags:
          - Variable List
        summary: List of used variables on this PANDEM-2 instance
        description: List variables published on this PANDEM-2 instace
        operationId: getVariableList
        parameters:
        responses:
            '200':
              description: List of variables and mofifiers
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/VariableListModel'
                application/xml:
                  schema:
                    $ref: '#/components/schemas/VariableListModel'
                text/plain:
                  schema:
                    type: string
        """


        var_path = util.pandem_path("files", "variables")
        var_paths = self.storage_proxy.list_files(var_path, include_files = False, include_dirs = True, recursive = False).get()
        var_dic = self.variables_proxy.get_variables().get()
        
        pub_vars = {v['path'] for v in var_paths }

        sources_path = util.pandem_path("files", "source-definitions")
        dls_paths = self.storage_proxy.list_files(sources_path).get()
        dlss = [self.storage_proxy.read_file(f['path']).get() for f in dls_paths if f['path'].endswith(".json")]
        used_vars = {col["variable"] for dls in dlss for col in dls['columns'] if "variable" in col}
        found_vars = {}
        for v in var_dic:
          if v in used_vars or v in pub_vars: 
             found_vars[v] = var_dic[v].copy()
             found_vars[v].pop("aliases")
             found_vars[v]["base_variable"] = found_vars[v]["variable"] if  found_vars[v]["variable"] != v else None
             found_vars[v]["variable"] = v
        ret = list(found_vars.values())
        ret.sort(key = lambda v: v["data_family"]+"_"+v["variable"])
        response = {'variables':ret}
        self.write(response)


class TimeSerieHandler(tornado.web.RequestHandler):
    def initialize(self, storage_proxy, variables_proxy):
        self.storage_proxy = storage_proxy
        self.variables_proxy = variables_proxy
    def post(self):
        """
        ---
        tags:
          - Time series
        summary: Retrieve infromation about time series avaiable in this PANDEM-2 instance
        description: Time series available in PANDEM-2
        operationId: getTimeSeries
        parameters:
        responses:
            '200':
              description: List of time series available
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/TimeSerieModel'
                application/xml:
                  schema:
                    $ref: '#/components/schemas/TimeSerieModel'
                text/plain:
                  schema:
                    type: string
        """
        storage_proxy = self.storage_proxy
        variables_proxy = self.variables_proxy
        var_dic = variables_proxy.get_variables().get()
        query = tornado.escape.json_decode(self.request.body)
        if query is not None:
          ts = variables_proxy.get_timeseries().get()
          comb = [(k, v) for (k,v) in query.items() if v is not None and k != 'indicator' and k != 'source']
          comb.sort(key = lambda v: v[0])
          source = query["source"]
          indicator = query["indicator"]
          variable = var_dic[indicator]["variable"]
          datevars = [v for v, varinfo in var_dic.items() if varinfo['type']=='date' and varinfo['variable']==v]

          data = variables_proxy.lookup([variable], source = source, combinations = [tuple(comb)], filter = {d:None for d in datevars}).get()
          # making an aggregation at day level
          resp = {}
          for key, values in data.items():
            for var, tuples in values.items():
              for t in tuples:
                value = t["value"]
                for datevar, dtime in t["attrs"].items():
                  date = dtime[0:10]
                  if (date, datevar) not in resp:
                    resp[(date, datevar)] = {'date':date, 'date_var':datevar, "key":json.dumps({k:v for k,v in key})}
                    #resp[(date, datevar)].update({k:v for k,v in keys})
                  if var not in resp[(date, datevar)]:
                    resp[(date, datevar)]["indicator"] = indicator
                    resp[(date, datevar)]["value"] = value
                  else :
                    # TODO: change the aggregation function depending on the unit
                    resp[(date, datevar)]["value"] = resp[(date, datevar)]["value"] + value
                    
        response = {"timeserie":list(resp.values())}
        self.write(response)

class TimeSeriesHandler(tornado.web.RequestHandler):
    def initialize(self, storage_proxy, variables_proxy):
        self.storage_proxy = storage_proxy
        self.variables_proxy = variables_proxy
        self._timeseries_hash = ''
        self._timeseries_cache = ''
    def get(self):
        """
        ---
        tags:
          - Time series
        summary: Retrieve data of a particular time serie defined by its combination filer passed as json port
        description: Get time serie data
        operationId: getTimeSerie
        parameters:
        responses:
            '200':
              description: Data of time serie
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/TimeSeriesModel'
                application/xml:
                  schema:
                    $ref: '#/components/schemas/TimeSeriesModel'
                text/plain:
                  schema:
                    type: string
        """
        storage_proxy = self.storage_proxy
        variables_proxy = self.variables_proxy
        # If no changes have been performed on timeseries and we have a let's try to reuse last value
        if self._timeseries_cache != '' and self._timeseries_hash != '' and self._timeseries_hash == self.variables_proxy.timeseries_hash.get():
            self.write(self._timeseries_cache)
        else:
            # recalculating time series
            dlss = [storage_proxy.read_file(f["path"]).get() for f in storage_proxy.list_files(util.pandem_path("files", "source-definitions")).get() if f["name"].endswith(".json")]
            source_names = {dls["scope"]["source"]:dls["scope"]["tags"][0] if "tags" in dls["scope"] and len(dls["scope"]["tags"]) > 0 else dls["scope"]["name"] for dls in dlss}
            reference_users = {dls["scope"]["source"]:dls["scope"]["reference_user"] if "reference_user" in dls["scope"] else "" for dls in dlss}
            data_quality = {dls["scope"]["source"]:dls["scope"]["data_quality"] if "data_quality" in dls["scope"] else "" for dls in dlss}
            source_descriptions = {dls["scope"]["source"]:dls["scope"]["source_description"] if "source_description" in dls["scope"] else "" for dls in dlss}
            var_dic = variables_proxy.get_variables().get()
            
            # getting referential aliases
            ref_labels = {code:k for k, varinfo in var_dic.items() if varinfo["type"] == "referential_label" for code in varinfo["linked_attributes"]}
            ref_attrs = {}
            ref_link = {}
            for k, varinfo in var_dic.items(): 
              if varinfo["type"] in ["characteristic", "referential_parent" ] and varinfo['linked_attributes'] is not None: 
                for code in varinfo["linked_attributes"]:
                  if code not in ref_attrs:
                    ref_attrs[code]= []
                  ref_attrs[code].append(k)
                  ref_link[k]=code
                  
            code_labels = {}
            code_attrs = {}
            no_labels = ["indicator__description", "source__reference_user", "source__source_description"]
            ts = variables_proxy.get_timeseries().get()
            self._timeseries_hash = variables_proxy.timeseries_hash.get()
            ts_values = [{k:v for k, v in key} for key in ts.keys()]
            # Adding nice to have information on time series
            for values in ts_values:
              # Adding related attribute characteristics
              for var, value in list(values.items()):
                if var in ref_attrs:
                  # Loading referential attrs if present 
                  for attr in  ref_attrs[var]:
                    if attr not in code_attrs:
                       link = variables_proxy.get_referential(var).get()
                    if link is not None:
                      code_attrs[attr] = {t['attr'][var]:t['attrs'][attr] for t in link if 'attrs' in t and 'attr' in t and attr in t['attrs'] and var in t['attr']}
                    # Taking label from referential
                    if attr in code_attrs and value in code_attrs[attr]:
                      values[f"ref__{attr}"] = code_attrs[attr][value]

              # Adding frienly labels
              for var, value in list(values.items()):
                var = var.split("__")[-1]
                if var in ref_labels:
                  # if variable is a characteristic with a link to a referential we have to use the referential as query table
                  if var not in ref_link:
                    label_name = f"{var}_label"
                    ref_var = var 
                  else:
                    ref_var = ref_link[var]
                    label_name = f"ref__{var}_label"
                  
                  # Loading referential labels if present 
                  label = ref_labels[ref_var]
                  if ref_var not in code_labels:
                    labels = variables_proxy.get_referential(label).get()
                    if labels is not None:
                      code_labels[ref_var] = {t['attrs'][ref_var]:t['attr'][label] for t in labels if 'attrs' in t and 'attr' in t and label in t['attr'] and ref_var in t['attrs']}
                  # Taking label from referential if exists
                  if ref_var in code_labels and value in code_labels[ref_var]:
                    values[label_name] = code_labels[ref_var][value]
                  # Taking code as value if not Nont 
                  elif value is not None:
                    values[label_name] = str(value) 
                  else:
                    values[label_name] = None
                  # Making case correction when possible
                  #if values[label_name] is not None and len(values[label_name])>2: #and ref_var not in no_labels:
                  #  values[label_name] = " ".join([(word[0].upper()+word[1:].lower() if len(word)>2 else word)  for word in re.split("_| |\\-", values[label_name])])

              # Adding indicator associated values
              if "indicator" in values:
                values["indicator__family"] = var_dic[values["indicator"]]["data_family"]
                values["indicator__description"] = var_dic[values["indicator"]]["description"]
                values["indicator__unit"] = var_dic[values["indicator"]]["unit"]
              # Adding source (DLS) associated values
              if "source" in values:
                values["source__table"] = values["source"]
                values["source__source_name"] = source_names[values["source"]]
                values["source__reference_user"] = reference_users[values["source"]]
                values["source__source_description"] = source_descriptions[values["source"]]
                values["source__data_quality"] = data_quality[values["source"]]


            response = {"timeseries":ts_values}
            self._timeseries_cache = response
            self.write(response)

class DatasetHandler(tornado.web.RequestHandler):
    def initialize(self, storage_proxy, variables_proxy):
        self.storage_proxy = storage_proxy
        self.variables_proxy = variables_proxy

    def post(self):
        """
        ---
        tags:
          - Get Data Set
        summary: Retrieve infromation about time series avaiable in this PANDEM-2 instance
        description: specific dataset if available in PANDEM-2
        operationId: getDataSet
        parameters:
        responses:
            '200':
              description: Returns a specific dataset if available
              content:
                application/json:
                  schema:
                    $ref: '#/components/schemas/DatasetModel'
                application/xml:
                  schema:
                    $ref: '#/components/schemas/DatasetModel'
                text/plain:
                  schema:
                    type: string
        """
        storage_proxy = self.storage_proxy
        variables_proxy = self.variables_proxy
        var_dic = variables_proxy.get_variables().get()
        
        query = tornado.escape.json_decode(self.request.body)
        if query is not None:
            ts = variables_proxy.get_timeseries().get()
            needed_variables, query = self.__build_needed_variables(query, var_dic)
            filtered_ts = self.__filter_ts(query, needed_variables, ts)

            # Getting columns of the dataframe
            data = {}
            columns = {'source'}
            for ((source, indicator), comb) in filtered_ts.items():
                d = variables_proxy.lookup([var_dic[indicator]['variable']], source=source,
                                          combinations=comb, filter={'reporting_period': None}).get()
                data[(source, indicator)] = d
                columns.add(indicator)

                # Identifying columns from data keys
                for ts_key in d.keys():
                    for k, v in ts_key:
                        columns.add(k)
                # Retrieving columns from nested attrs
                for ts_dict in d.values():
                    for ts_values in ts_dict.values():
                        if len(ts_values) > 0 and 'attrs' in ts_values[0]:
                            columns.update(ts_values[0]['attrs'].keys())

            # Filling the dataframe with values
            rows = {}
            for ((source, indicator), v) in data.items():
                for ts_key, ts_dict in v.items():
                    for ts_values in ts_dict.values():
                        for ts_value in ts_values:
                            attrs_date_key = list(ts_value['attrs'].keys())[0]
                            if var_dic[attrs_date_key]['type'] == 'date':
                                date_element = (
                                    attrs_date_key, ts_value['attrs'][attrs_date_key])
                                row_key = list(ts_key)
                                row_key.append(date_element)
                                row_key = tuple(row_key)
                                if row_key not in rows:
                                    rows[row_key] = self.__init_row(
                                        ts_key, columns, date_element, source)
                            rows[row_key] = self.__update_indicator(
                                ts_value, rows, row_key, date_element, indicator)

            df = pd.DataFrame.from_dict(rows, orient='index', dtype=None)
            df = df.replace({np.nan: None})
            self.write({"dataset": df.to_dict('records')})


    def __build_needed_variables(self, query, var_dic):
        """Build a list of needed_variables and adds required filters"""
        needed_variables = []
        for el in query['select']:
            if var_dic[el]['type'] not in ['observation', 'indicator', 'resource', 'date']:
                needed_variables.append(var_dic[el]['variable'])
            for modifier in var_dic[el]['modifiers']:
                query['filter'][modifier['variable']] = modifier['value']

        needed_variables.extend(query['filter'].keys())
        return list(set(needed_variables)), query
    
    def __filter_ts(self, query, needed_variables, ts):
        """Keep the timeseries which have both needed_variables and proper filters"""
        filtered_ts = {}

        for k, v in ts.items():
            is_in_filter, is_in_needed = True, True
            for nv in needed_variables:
                if nv not in list(map(lambda v: v[0], k)):
                    is_in_needed = False
                    break
            if is_in_needed:
                for k2, v2 in query['filter'].items():
                    if k2 not in list(map(lambda x: x[0], k)):
                        is_in_filter = False
                        break
                    if v2 not in list(map(lambda x: x[1], k)):
                        is_in_filter = False
                        break
            if is_in_filter and is_in_needed:
                comb = [(i, j) for (i, j) in k if j is not None and i !=
                        'indicator' and i != 'source']
                comb.sort(key=lambda v: v[0])

                source = [(i, j)
                          for (i, j) in k if j is not None and i == 'source'][0][1]
                indicator = [(i, j)
                            for (i, j) in k if j is not None and i == 'indicator'][0][1]
                if (source, indicator) not in filtered_ts:
                    filtered_ts[(source, indicator)] = set()
                filtered_ts[(source, indicator)].add(tuple(comb))
        return filtered_ts
    
    def __init_row(self, ts_key, columns, date_element, source):
        """Initialize dictionary's columns and add value if not indicator"""
        row = {}
        for col in columns:
            tuples_keys = list(map(lambda x: x[0], ts_key))
            tuples_values = list(map(lambda x: x[1], ts_key))
            if col in tuples_keys:
                row[col] = tuples_values[tuples_keys.index(col)]
            else:
                row[col] = None
        row[date_element[0]] = date_element[1]
        row['source'] = source
        return row
    
    def __update_indicator(self, ts_value, rows, row_key, date_element, indicator):
        """Updates a dictionary with all indicators values"""
        if ts_value['attrs']:
            for i, j in ts_value['attrs'].items():
                if i == date_element[0]:
                    rows[row_key][indicator] = ts_value['value']
                    return rows[row_key]


@components.schemas.register
class SourcesModel(object):
    """
    ---
    type: object
    description: Sources model representation
    properties:
        sources:
            type: array
    """

@components.schemas.register
class JobsModel(object):
    """
    ---
    type: object
    description: Sources model representation
    properties:
        Jobs:
            type: array
    """

@components.schemas.register
class IssueModel(object):
    """
    ---
    type: object
    description: Issue model representation
    properties:
        issues:
            type: array
    """

@components.schemas.register
class SourceDetailModel(object):
    """
    ---
    type: object
    description: Source detail model representation
    properties:
        sources:
            type: array
        definitions:
            type: dict
    """

@components.schemas.register
class VariableListModel(object):
    """
    ---
    type: object
    description: Variable List model representation
    properties:
        variables:
            type: array
    """

@components.schemas.register
class TimeSeriesModel(object):
    """
    ---
    type: array
    description: List of variables characterizing all time series
    properties:
        timeseries:
            type: array
    """

@components.schemas.register
class DatasetModel(object):
    """
    ---
    type: array
    description: Dataset characterizing a time serie
    properties:
        dataset:
            type: array
    """


class Application(tornado.web.Application):

    def __init__(self, storage_proxy, variables_proxy, pipeline_proxy):
        settings = {"debug": False}
        self._routes = [
          tornado.web.url(r"/jobs", JobsBySourceHandler, {'storage_proxy': storage_proxy, 'pipeline_proxy':pipeline_proxy}),
          tornado.web.url(r"/sources", SourcesHandler, {'storage_proxy': storage_proxy}),
          tornado.web.url(r"/issues", IssuesHandler, {'storage_proxy': storage_proxy}),
          tornado.web.url(r"/source_details", SourceDetailsHandler, {'storage_proxy': storage_proxy}),
          tornado.web.url(r"/variable_list", VariableListHandler, {'storage_proxy': storage_proxy, 'variables_proxy':variables_proxy}),
          tornado.web.url(r"/timeseries", TimeSeriesHandler, {'storage_proxy': storage_proxy, 'variables_proxy':variables_proxy}),
          tornado.web.url(r"/timeserie", TimeSerieHandler, {'storage_proxy': storage_proxy, 'variables_proxy':variables_proxy}),
          tornado.web.url(r"/dataset", DatasetHandler, {'storage_proxy': storage_proxy, 'variables_proxy':variables_proxy})
        ]
        setup_swagger(
            self._routes,
            swagger_url="/",
            description="",
            api_version="1.0.0",
            title="Pandem source API"
        )
        super(Application, self).__init__(self._routes, **settings)



import os
from . import worker
from abc import ABC, abstractmethod, ABCMeta
import logging
import functools
import requests
import json
import re
import copy
import itertools

l = logging.getLogger("pandem-nlp")

class NLPAnnotator(worker.Worker):
    __metaclass__ = ABCMeta  
    def __init__(self, name, orchestrator_ref, settings, run_nlp): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)    
        self.run = run_nlp
        if self.run:
          self._models_path = settings["pandem"]["source"]["nlp"]["models_path"]
          self._tf_port = settings["pandem"]["source"]["nlp"]["tensorflow_server_port"]
          self._tf_url = f"{settings['pandem']['source']['nlp']['tensorflow_server_protocol']}://{settings['pandem']['source']['nlp']['tensorflow_server_host']}:{settings['pandem']['source']['nlp']['tensorflow_server_port']}"
          self._tf_version = settings["pandem"]["source"]['nlp']["tensorflow_server_version"]
          self._model_categories = settings["pandem"]["source"]["nlp"]["categories"]
          self._model_languages = settings["pandem"]["source"]["nlp"]["languages"]
    def on_start(self):
        super().on_start()
        if self.run:
          self._storage_proxy = self._orchestrator_proxy.get_actor('storage').get().proxy()
          self._pipeline_proxy = self._orchestrator_proxy.get_actor('pipeline').get().proxy() 
          self._variables_proxy = self._orchestrator_proxy.get_actor('variables').get().proxy()
          #self._models = self.get_models()
          self._models = None

    def annotate(self, list_of_tuples, path, job):
      if self.run:
        # getting tuples from cache
        list_of_tuples = list_of_tuples.value() 
        if self._models is None:
            self._models = self.get_models()
        # gathering information about nlp categories
        endpoints = self.model_endpoints()
        categories = {m:self._model_categories[m] for m in endpoints.keys() if (m in self._model_categories) }
        
        #gatherint information for geo annotation
        variables = self._variables_proxy.get_variables().get()
        geos = {var["variable"] for var in variables.values() if var["type"] == "geo_referential"}
        alias_vars = {
          var["variable"]:var["linked_attributes"][0] 
          for var in variables.values() 
          if var["type"] in ["referential_alias", "referential_label"] and var["linked_attributes"] is not None and var["linked_attributes"][0] in geos
        }
        aliases = {}
        for alias_var, code_var in alias_vars.items():
          alias_values = self._variables_proxy.read_variable(alias_var, {}).get()
          if alias_values is not None:
            alias_map = {t["attr"][alias_var].lower():t["attrs"][code_var] for t in alias_values if "attr" in t and "attrs" in t and alias_var in t["attr"] and code_var in t["attrs"]} 
            if code_var not in aliases:
              aliases[code_var] = alias_map
            else :
              aliases[code_var].update(alias_map)
        
        alias_regex = {code_var:re.compile('|'.join([f"\\b{re.escape(alias)}\\b" for alias in aliases[code_var]])) for code_var in aliases}

        text_field = "article_text"
        lang_field = "article_language"
     
        
        l.debug("Evaluatig NLP models")
        annotated = []
        count = 0
        for lang in self._model_languages:
          for to_annotate in self.slices((t for t in list_of_tuples['tuples'] if "attrs" in t and text_field in t["attrs"] and lang_field in t["attrs"] and t["attrs"][lang_field]==lang), 10000):
            # Annotating using tensorflow categories
            for m in categories.keys():
              if m in self._model_languages[lang]: 
                texts = (t["attrs"][text_field] for t in to_annotate)
                data = json.dumps({"instances": [[t] for t in texts]})
                result = requests.post(f"{endpoints[m]}:predict", data = data, headers = {'content-type': "application/json"}).content
                annotations = json.loads(result)["predictions"]
                for t, pred in zip(to_annotate, annotations):
                  best = functools.reduce(lambda a, b: a if a[1]>b[1] else b, enumerate(pred))[0]
                  at = copy.deepcopy(t)
                  # creating a new tuple with the best category for this model
                  at["attrs"][f"article_cat_{m}"] = categories[m][best]
                  annotated.append(at)
                  # updating exitsting tuple with category ALL
                  t["attrs"][f"article_cat_{m}"] = "All"
            count = count + len(to_annotate)
            l.debug(f"{count} articles annotated")
        l.debug("Evaluating geolocation")

        # Annotating geographically using extra simplistic approach
        count = 0
        geo_annotated = []
        geo_annotation = {}
        to_annotate = [ t for t in list_of_tuples['tuples'] if "attrs" in t and text_field in t["attrs"]]
        for geo_var, regex in alias_regex.items():
          texts = [t["attrs"][text_field] for t in to_annotate]
          for t in to_annotate + annotated:
            if not geo_var in t["attrs"]:
              text = t["attrs"][text_field].lower()
              if text not in geo_annotation:
                match = re.search(alias_regex[geo_var], text)
                geo_annotation[text] = match
              else :
                match = geo_annotation[text]

              if match is not None:
                matched_alias = match.group()
                at = copy.deepcopy(t)
                at["attrs"][geo_var] = aliases[geo_var][matched_alias]
                geo_annotated.append(at)
              t["attrs"][geo_var] = "All"
            count = count + 1
            if count % 10000 == 0:
              l.debug(f"{count} articles geo annotated")
        l.debug(f"{count} articles geo annotated")
        l.debug("Removing language from tuples")
        
        # Adding annotated tuples
        list_of_tuples['tuples'].extend(annotated)
        list_of_tuples['tuples'].extend(geo_annotated)
        
        # Removing language field since it is not interesting for generating separated time series
        for t in list_of_tuples['tuples']:
          if 'attrs' in t and lang_field in t['attrs']:
            t['attrs'].pop(lang_field)
              
        ret = self._storage_proxy.to_job_cache(job["id"], f"std_{path}", list_of_tuples).get()
        self._pipeline_proxy.annotate_end(ret, path = path, job = job)

    def slices(self, iterable, size):
       head = list(itertools.islice(iterable, size))
       while len(head) > 0:
         yield head
         head = list(itertools.islice(iterable, size))

    def get_models(self):
      if os.path.exists(self._models_path):
        return list(filter(lambda v: not v.startswith("."), next(os.walk(self._models_path))[1]))
      else: 
        raise FileNotFoundError(f"Cannot find the NLP models folder {self._models_path}")

    def model_endpoints(self):
      return {m:f"{self._tf_url}/v1/models/{m}" for m in self.get_models()}

       




import os
import time
from datetime import datetime
import threading
from . import worker
from . import util
from abc import ABC, abstractmethod, ABCMeta
import pandas as pd
import numpy as np
import json
import isoweek
import logging
import re
l = logging.getLogger("pandem.dfreader")

class DataframeReader(worker.Worker):
    __metaclass__ = ABCMeta

    def __init__(self, name, orchestrator_ref, storage_ref, settings):
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)


    def on_start(self):
        super().on_start()
        self._variables_proxy=self._orchestrator_proxy.get_actor('variables').get().proxy()
        self._pipeline_proxy=self._orchestrator_proxy.get_actor('pipeline').get().proxy()
        self._storage_proxy = self._orchestrator_proxy.get_actor('storage').get().proxy()


    def get_variables(self):
        return self._variables_proxy.get_variables().get()

    def check_df_columns(self, df, job, dls, file_name, variables):
        """Checks if the DLS columns names exist in dataframe
        and if the variable indicated in the column element from dls exists in variables.json files
        and if yes, checks the dataframe variables types, if not as expected in variables.json, raises an issue"""
        var=variables
        dls_col_list=[]
        issues = {}

        if 'columns' in dls.keys() :
            dls_col_list = dls['columns']
        else :
            message=("DLS file not conform, 'columns' element is missing.")
            issue={ 'step': job['step'],                       
                    'line' : 0,                        
                    'source' : dls['scope']['source'],
                    'file' : file_name,
                    'message' : message,
                    'raised_on' : datetime.now(),
                    'job_id' : job['id'],
                    'issue_type' : "dls-not-conform",
                    'issue_severity':"error"}
            issues["DLS file"] = issue

        
        for item in dls_col_list:
            if "name" not in item: 
              message=("DLS file not conform, 'columns' must contain a list of objects containint elements 'name'")
              issue={ 'step': job['step'],                       
                    'line' : 0,                        
                    'source' : dls['scope']['source'],
                    'file' : file_name,
                    'message' : message,
                    'raised_on' : datetime.now(),
                    'job_id' : job['id'],
                    'issue_type' : "dls-not-conform",
                    'issue_severity':"error"
                    }
              issues["DLS file"] = issue
            elif item['name'] not in df.columns :
                message=(f"Column '{item['name']}' not found in dataframe ")
                issue={ 'step' : job['step'],
                        'line' : 0,
                        'source' : dls['scope']['source'],
                        'file' : file_name,
                        'message': message,
                        'raised_on' : datetime.now(),
                        'job_id' : job['id'],
                        'issue_type' : "col-not-found",
                        'issue_severity':"error"}                       
                issues[item["name"]] = issue
            elif 'variable' in item and item['variable'] in var:
                if var[item['variable']]['unit'] == 'date':
                    if df[item['name']].dtypes in ['date', 'object', 'str', 'datetime64[ns]', 'int', 'int64'] :
                        issues[item["name"]] = None
                    else :
                        message = (f"Type '{df[item['name']].dtypes}' in source file is not compatible with variable {item['variable']} with unit 'date' ")
                        issue={ 'step' : job['step'],
                                'line' : 0,
                                'source' : dls['scope']['source'],
                                'file' : file_name,
                                'message' : message,
                                'raised_on' : datetime.now(),
                                'job_id' : job['id'],
                                'issue_type' : "ref-not-found",
                                'issue_severity':"warning"}
                        issues[item["name"]] = issue
                elif self.is_numeric_unit(var[item['variable']]['unit']):
                    if df[item['name']].dtypes in ['integer', 'str', "object", "float", "float64", "int64", 'int'] :
                        issues[item["name"]] = None
                    else :
                        message = (f"Type '{df[item['name']].dtypes}' in source file is not compatible with variable {item['variable']} with unit 'integer' ")
                        issue = {'step' : job['step'],
                                'line' : 0,
                                'source' : dls['scope']['source'],
                                'file' : file_name,
                                'message' : message,
                                'raised_on' : datetime.now(),
                                'job_id' : job['id'],
                                'issue_type' : "ref-not-found",
                                'issue_severity':"warning"}
                        issues[item["name"]] = issue

                elif var[item['variable']]['unit'] == None :
                    issues[item["name"]] = None
                elif var[item['variable']]['unit'] in ['string', 'range'] :
                    issues[item["name"]] = None
            elif "variable" in item:
                message = (f"Variable {item['variable']} defined on source definition file is unknown")
                issue = {'step' : job['step'],
                        'line' : 0,
                        'source' : dls['scope']['source'],
                        'file' : file_name,
                        'message' : message,
                        'raised_on' : datetime.now(),
                        'job_id' : job['id'],
                        'issue_type' : "ref-not-found",
                        'issue_severity':"error"
                }
                issues[item["name"]] = issue
        return issues


    def translate(self, value, dtype, unit, format = {}):   # parameters to pass after : df, dls, file_name
        """Convert the format of dataframe column if expected unit is different"""
        if pd.isna(value):
            return None
        if unit is None or unit in ["str"] :
            if dtype in ["str"]:
                return value
            else:
                return str(value)
        elif unit == 'date':
            if np.issubdtype(dtype, np.datetime64):
                return value
            elif type(value) == datetime:
                return value
            elif format.get("date_format") is None:
                raise AssertionError("You have to provide a format to parse a date on Data source definition, on ['acquisition']['format']['date_format']")
            elif format.get("date_format") =='isoweek':
                if dtype in ["int", "int64", "float", "float64"]:
                    return isoweek.Week(int(int(value)/100),int(value)%100).thursday()
                else:
                    year = [int(v) for v in re.findall("[0-9]+", value) if int(v) > 100][0]
                    week = [int(v) for v in re.findall("[0-9]+", value) if int(v) < 100][0]
                    return isoweek.Week(year,week).thursday()
            else:
                return datetime.strptime(value, format.get('date_format'))
        elif self.is_numeric_unit(unit) :
            if dtype in ["int", "int64", "float", "float64"]: #, "object"
              if np.isnan(value):
                return None
              else:
                return value
            else:
                value = str(value).replace(" ","")
                if "thousands_separator" in format and format["thousands_separator"]!=".":
                  value = value.replace(format["thousands_separator"], "") 
                if "decimal_separator" in format and format["decimal_separator"]!=".":
                  value = value.replace(format["decimal_separator"], ".") 
                return float(value)
        elif unit=='range':
            return self.transform_range(value)
        else: 
          raise ValueError("undefined case for casting")
    

    def df2tuple(self, df, path, job, dls):
        variables = self.get_variables()
        # getting df from cache
        df = df.value()
        dls_col_list=[]
        file_name = util.absolute_to_relative(path, "files/staging/")

        # Checking column names and compatible types
        col_issues = self.check_df_columns(df = df, job = job, dls = dls, file_name = path, variables = variables)
        issues = list([i for i in col_issues.values() if i is not None]) 
        types_ok = dict((t["name"],  t["name"] in col_issues and col_issues[t["name"]] is None) for t in  dls['columns'])
        col_vars = dict((t["name"], t["variable"]) for t in  dls['columns'] if 'variable' in t)
        formats = {v:self.format_for(v, dls, variables) for v in col_vars.values()}
        col_types = dict((k,variables[v]["type"]) for k,v in col_vars.items() if v in variables)
        insert_cols = dict((t["name"], ("obs" if col_types[t["name"]] in ["observation", "indicator", "resource"] else "attr")) 
          for t in  dls['columns'] 
          if types_ok[t["name"]] and
            'variable' in t and
            ("action" in t and t["action"] == "insert" or col_types[t["name"]] in ["observation", "indicator", "resource"])
        )
        attr_cols = {}
        for col in insert_cols.keys():
          typ = col_types[col]
          if typ in ["observation", "indicator", "resource"]:
            attr_cols[col] =  [k for k, v in col_types.items() if v not in  ["observation", "indicator", "resource"] and types_ok[k]]
          else:
            attr_cols[col] =  [k for k, v in col_vars.items() if types_ok[k] and v not in  ["observation", "indicator", "resource"] and k != col]
        tuples = []
        for row in range(len(df)):
            for col, group in insert_cols.items():
                tup = {
                  "attrs":{
                    "line_number":df["line_number"][row]
                    ,"source":dls["scope"]["source"]
                    ,"file":file_name
                  },
                }
                self.translate_or_issue(
                    tup = tup,
                    group = group,
                    val = df[col][row], 
                    format = formats[col_vars[col]],
                    issues = issues, 
                    dtype = df[col].dtypes, 
                    unit = variables[col_vars[col]]["unit"], 
                    dls = dls, 
                    job = job, 
                    line_number = df["line_number"][row], 
                    file_name = file_name,
                    var_name = col_vars[col],
                    variables = variables
                )
                for attr_col in attr_cols[col]:
                    self.translate_or_issue(
                        tup = tup,
                        group = "attrs",
                        val = df[attr_col][row], 
                        format = formats[col_vars[attr_col]],
                        issues = issues, 
                        dtype = df[attr_col].dtypes, 
                        unit = variables[col_vars[attr_col]]["unit"], 
                        dls = dls, 
                        job = job, 
                        line_number = df["line_number"][row], 
                        file_name = file_name,
                        var_name = col_vars[attr_col],
                        variables = variables
                    )
                tuples.append(tup)
        ret = {"scope":dls["scope"].copy()}
        ret["scope"]["file_name"] = file_name
        # validating that globals are property instantiated
        #if "scope" in dls and "globals" in dls["scope"] : 
        #  ret["scope"]["globals"] = self.add_values(dls["scope"]["globals"], tuples = [], issues = issues, dls = dls, job = job, file_name = file_name)
        # instantiating update scope based on existing values on tuple
        if "scope" in dls and "update_scope" in dls["scope"]:
          ret["scope"]["update_scope"] = self.add_values(dls["scope"]["update_scope"], tuples = tuples, variables = variables, issues = issues, dls = dls, job = job, file_name = file_name)
        ret["tuples"] = tuples

        ret = self._storage_proxy.to_job_cache(job["id"], f"tup_{path}", ret).get()
        self._pipeline_proxy.read_df_end(tuples = ret, n_tuples = len(tuples), issues = issues, path = path, job = job)
    

    def add_values(self, var_list, tuples, variables, issues, dls, job, file_name):
        ret = []
        for var in var_list:
          if "variable" not in var:
            pass
          else :
            # Getting base varabe if a derivated variabe
            var_name = var["variable"]
            base_var = variables[var_name]["variable"]
            r = {"variable": base_var}
            modifiers  = self.add_alias_context({"attrs":{}}, var_name, variables)
            for k, v in modifiers["attrs"].items():
              ret.append({"variable":k, "value":v})
            
            if "value" in var :
              # all good no need to get distinct values just making sure the value is put as an array
              if not isinstance(var["value"], list):
                r["value"] = [var["value"]]
              else :
                r["value"] = var["value"]
            else :
              # value is missing we are going to try to get it from tuples
              r["value"] = list(set(x["attrs"][base_var] for x in tuples if base_var in x["attrs"]))
            ret.append(r)

            if len(r["value"])==0:
              message = (f"Variable {var['variable']} needs to be instantiated ad per DLS but no values where found on dataset")
              issue = {'step' : job['step'],
                      'line' : 0,
                      'source' : dls['scope']['source'],
                      'file' : file_name,
                      'message' : message,
                      'raised_on' : datetime.now(),
                      'job_id' : job['id'],
                      'issue_type' : "cannot-instantiate",
                      'issue_severity':"warning"
              }
              issues.append(issue)
        return ret
    def translate_or_issue(self, tup, group, val, format, issues, dtype, unit, dls, job, line_number, file_name, var_name, variables):
        trans = None
        try:
          trans = self.translate(val, dtype, unit, format)
        except AssertionError as e:
          issues.append({
            "step":job['step'],
            "line":line_number,
            "source":dls['scope']['source'],
            "file":file_name,
            "message":f"Cannot cast value {val} into unit {unit} as {dtype}\n {e}",
            "raised_on":datetime.now(),
            "job_id":job['id'],
            "issue_type":"cannot-cast",
            'issue_severity':"error"
          })
        except Exception as e:
          issues.append({
            "step":job['step'],
            "line":line_number,
            "source":dls['scope']['source'],
            "file":file_name,
            "message":f"Cannot cast value {val} into unit {unit} as {dtype}\n {e}",
            "raised_on":datetime.now(),
            "job_id":job['id'],
            "issue_type":"cannot-cast",
            'issue_severity':"warning"
          })
        if trans is not None:
            if group not in tup:
              tup[group] = {}
            tup[group][variables[var_name]["variable"]] = trans
            self.add_alias_context(tup, var_name, variables)

    def add_alias_context(self, tup, var_name, variables):
        if var_name in variables and "modifiers" in variables[var_name]:
            for mod in variables[var_name]["modifiers"]:
              mod_var = mod["variable"]
              if not "attrs" in tup:
                tup["attrs"] = {}
              if mod_var not in tup["attrs"] or tup["attrs"][mod_var] is None:
                tup["attrs"][mod_var] = mod["value"]
        return tup

    def transform_range(self, value):
      nb = re.findall('[0-9]+', value)
      nb_size = len(nb)
      if nb_size == 0:
        return None
      elif nb_size == 1:
        return f'-{nb[0]}'if int(nb[0]) <= 40 else f'{nb[0]}-'
      elif nb_size == 2:
        nb = sorted(list(map(int, nb)))
        return f'{nb[0]}-{nb[1]}'
      else:
        raise ValueError(f'Too much numbers in list ({nb_size}): {nb}')

    def is_numeric_unit(self, unit):
      if unit is None:
        return False 

      unit = unit.lower().replace(" ", "")
      if unit in ['people', 'number', 'qty', 'days', 'people/people', 'units/time']:
        return True
      return False

    def format_for(self, var_name, dls, variables):
      col_def = list([col for col in dls["columns"] if "variable" in col and col["variable"]==var_name])
      format = {}
      for attr in ["decimal_sign", "thousands_separator", "date_format"]:
        if len(col_def) > 0 and attr in col_def[0]:
          format[attr] = col_def[0][attr]
        elif "acquisition" in dls and "format" in dls["acquisition"] and attr in dls["acquisition"]["format"]: 
          format[attr] =  dls["acquisition"]["format"][attr]
      return format



from . import acquisition
from . import util
import time
import json
from .util import JsonEncoder
import gzip
from datetime import datetime, timedelta
import urllib
import requests
import os
from math import inf
import re
import logging
import xml.etree.ElementTree
from email.utils import parsedate_to_datetime
import pytz

l = logging.getLogger("pandem.medysis")

class AcquisitionMedisys(acquisition.Acquisition):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings, channel = "twitter")
    
    def on_start(self):
        super().on_start()
    
    def add_datasource(self, dls, force_acquire):
      if len(self.current_sources) > 0:
        raise ValueError("Medisys aqquisition support only a singlr DLS, others will be ignored")
      if "acquisition" in dls and "channel" in dls["acquisition"] and "topics" in dls["acquisition"]["channel"]:
        self._topics  = dls["acquisition"]["channel"]["topics"].keys()
        self._maingroup  = dls["acquisition"]["channel"]["main_group"]
        self._topic_categories = {}
        self._included_regex = {}
        self._topic_groups = {}
        self._current_dir = util.pandem_path("files", "medisys", "current")
        self._arc_dir = util.pandem_path("files", "medisys", "archive")
        if not os.path.exists(self._current_dir):
          os.makedirs(name = self._current_dir)
        if not os.path.exists(self._arc_dir):
          os.makedirs(name = self._arc_dir)

        for topic in self._topics:
          self._included_regex[topic] = ""

          in_maingroup = dls["acquisition"]["channel"]["topics"][topic]["group"] == self._maingroup
          has_cat = False
          if "categories" in dls["acquisition"]["channel"]["topics"][topic]:
            self._topic_categories[topic] = set(dls["acquisition"]["channel"]["topics"][topic]["categories"])
            has_cat = True
          has_phrase = False
          if "phrases" in dls["acquisition"]["channel"]["topics"][topic]:
            phrases = dls["acquisition"]["channel"]["topics"][topic]["phrases"]
            self._included_regex[topic] = "|".join(map(lambda v: re.escape(v.lower()), phrases))  
            has_phrase = True
          if not has_phrase and not has_cat:
            raise ValueError("Medysis DLS topics need to have either  phrases property with a list of phrases to use or a categories property with a list of properties to follow")
          if in_maingroup and not has_cat:
            raise ValueError("Medysis DLS main topucs need to have a category")
          if "group" in dls["acquisition"]["channel"]["topics"][topic]:
            self._topic_groups[topic] =  dls["acquisition"]["channel"]["topics"][topic]["group"]
      else: 
        raise ValueError("If a Medysis DLS is found it should contain a (possible empty) list of topics under ['acquisition']['channel']")
      
      if "acquisition" in dls and "channel" in dls["acquisition"] and "languages" in dls["acquisition"]["channel"]:
        self._languages = [l.lower() for l in dls["acquisition"]["channel"]["languages"]]
      else:
        self._languages = None

      if "acquisition" in dls and "channel" in dls["acquisition"] and "excluded_phrases" in dls["acquisition"]["channel"]:
        self._excluded_regex =  "|".join(map(lambda v: re.escape(v.lower()), dls["acquisition"]["channel"]["excluded_phrases"]))
      else:
        self._excluded_regex = None

      super().add_datasource(dls, force_acquire)


    def new_files(self, dls, last_hash):
        # last_hash is the last until timespamp we captured
        if last_hash is None or last_hash == '':
          # If no last_hash is provided then 'until' = now and 'old_target' = a very early date
          until = datetime.strftime(datetime.utcnow(), "%Y-%m-%dT%H:%M:%SZ")
          old_target = "2000-00-00T00:00:00Z"
        else: 
          # If a hash is provided then then 'until' = now and 'old_target' = last_hash + 1sec
          until = datetime.strftime(datetime.utcnow(), "%Y-%m-%dT%H:%M:%SZ")
          old_target = datetime.strftime(datetime.strptime(last_hash, "%Y-%m-%dT%H:%M:%SZ") + timedelta(0,1), "%Y-%m-%dT%H:%M:%SZ")

        # looping over results until the endpoint produces no more articles
        # until will be the new hash after a successful iteration
        current_hash = until
        stop = False
        retries = 10
        retried = 0
        articles = []
        while(not stop):
          #Building the query url
          if self._languages is None:
            lang = "?"
          else:
            lang = f"?language={urllib.parse.quote(','.join(self._languages))}"
          dateto = f"dateto={urllib.parse.quote(until)}"
          datefrom = f"datefrom={urllib.parse.quote(old_target)}"
          categories = {v for vals in self._topic_categories.values() for v in vals}
          category = f"category={urllib.parse.quote(','.join(categories))}"
          url = f"https://medisys.newsbrief.eu/rss/{lang}&type=search&mode=advanced&{dateto}&{datefrom}&{category}"
          l.debug(url) 
          # getting the xml for the feed url
          res = requests.get(url)
          if res.status_code != 200:
            # request failed retrying on a arithmetic progression fashion
            l.warning(f"Medysis request failed with status {res.status_code} and reason: '{res.reason}'")
            if retried < retries:
              retried = retried + 1
              to_sleep = 10 * retried * retried
              l.info(f"Retrying again in {to_sleep} seconds")
              time.sleep(to_sleep)
            else:
              l.error("Cannot connect to the medisys endpoint, all retries has been exhausted")
              raise ConnectionError("Cannot connect to the medisys endpoint")
          else:
            # we have a succesfull connection we can parse the xml and get the articles
            tree = xml.etree.ElementTree.fromstring(res.content)
            nsmap = {"emm":"http://emm.jrc.it", "iso":"http://www.iso.org/3166", "georss":"http://www.georss.org/georss", "content":"http://purl.org/rss/1.0/modules/content/"} 
            new_art = []
            # Getting elements to obtain from tree
            if tree.find("channel") is not None:
              for item in tree.find("channel").findall("item"):
                art = {}
                new_art.append(art)
                # Getting single elements
                art["guid"] = [(node.text if node is not None else None) for node in [item.find("guid", nsmap)]][0]
                art["title"] = [(node.text if node is not None else None) for node in [item.find("title", nsmap)]][0]
                art["link"] = [(node.text if node is not None else None) for node in [item.find("link", nsmap)]][0]
                art["keywords"] = [(node.text if node is not None else None) for node in [item.find("emm:keywords", nsmap)]][0]
                art["pub_date"] = [(node.text if node is not None else None) for node in [item.find("pubDate", nsmap)]][0]
                art["lang"] = [(node.text if node is not None else None) for node in [item.find("iso:language", nsmap)]][0]
                art["article_count"] = 1

                # getting categories
                art["categories"] = [node.text for node in item.findall("category", nsmap)]
                
                # parsing the publication date to datetime
                art["pub_date"]= parsedate_to_datetime(art["pub_date"]).astimezone(pytz.utc)
                art["rep_date"]= datetime.strptime(datetime.strftime(art["pub_date"], "%Y-%m-%dT%H:00:00Z"), "%Y-%m-%dT%H:%M:%SZ")
 
                # getting full text of article (on this case is just the concatenation of the title and keywords
                art["text"] = f"{art['title']}. {art['keywords']}"

                # associating topics to the article
                for t in self.matching_topics(art["text"], art["categories"]):
                  if t in self._topic_groups:
                    col_name = f"topic_{self._topic_groups[t]}"
                  else:
                    col_name = "topic"
                  if col_name in art:
                    art[col_name].append(t)
                  else:
                    art[col_name] = [t]

          if len(new_art) > 0:
            # keep trying with older articles
            # new until will be the oldest of all articles minus one second
            l.debug(f"{len(new_art)} article obtained from medisys on a request")
            until = min(datetime.strftime(art["pub_date"] - timedelta(0,1), "%Y-%m-%dT%H:%M:%SZ") for art in new_art)
            articles.extend(new_art)
          else :
            # stop trying to get articles
            stop = True

        # collection finished, we can write the file 
        l.debug(f"{len(articles)} article obtained in total from medisys")
        
        
        # archiving previous files
        for to_arc in filter(lambda f: f.endswith(".json.gz"), os.listdir(self._current_dir)):
          os.rename(os.path.join(self._current_dir, to_arc), os.path.join(self._arc_dir, to_arc))

        # writing data to json.gz file
        fname = f"{datetime.now().strftime('%Y.%m.%d.%H.%M.%S')}.json.gz"
        path = os.path.join(self._current_dir, fname)
        with gzip.GzipFile(path, 'a') as fout:
          fout.write(json.dumps(articles, cls=JsonEncoder, indent = 4).encode('utf-8')) 

        return {"hash":current_hash, "files":[path]}  
          
    def matching_topics(self, text, categories):
        return list(filter(
          lambda topic: 
            (len(self._included_regex[topic]) > 0 and re.search(self._included_regex[topic], text.lower()) is not None and re.search(self._excluded_regex, text.lower()) is None)
            or (topic in self._topic_categories and not self._topic_categories[topic].isdisjoint(categories)),
          self._topics
        ))
        






# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

setup(
    name='pandem-source',
    version='1.0.9',
    description='Routinely collecting heterogeneous surveillance data into a coherent database',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Francisco Orchard',
    author_email='f.orchard@epiconcept.fr',
    url='https://github.com/pandem2/pandem-source',
    license="EUPL-1.2",
    install_requires=[
      "pyyaml",
      "pandas",
      "openpyxl",
      "pykka",
      "lxml",
      "isoweek",
      "tweepy",
      "tornado",
      "asyncio",
      "tornado-rest-swagger"
    ],
    packages=find_packages(exclude=('tests', 'docs')),
    include_package_data=True
)




import time
from . import worker
from abc import ABC, abstractmethod, ABCMeta
import logging as l
import numpy as np
import copy
import json
from .util import JsonEncoder

class Aggregator(worker.Worker):
  __metaclass__ = ABCMeta  
  def __init__(self, name, orchestrator_ref, settings): 
      super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)    

  def on_start(self):
      super().on_start()
      self._storage_proxy = self._orchestrator_proxy.get_actor('storage').get().proxy()
      self._pipeline_proxy = self._orchestrator_proxy.get_actor('pipeline').get().proxy() 
      self._variables_proxy = self._orchestrator_proxy.get_actor('variables').get().proxy()

  def aggregate(self, list_of_tuples, job):
    list_of_tuples = self.prepare_tuples(list_of_tuples)

    variables = self._variables_proxy.get_variables().get()
    cumm = {}
    untouched = []
    geos = {var["variable"] for var in variables.values() if var["type"] == "geo_referential"}
    geo_parents = {var["linked_attributes"][0]:var["variable"] for var in variables.values() if var["type"] == "referential_parent" and var["linked_attributes"] is not None and var["linked_attributes"][0] in geos}

    var_asc = {code:self.rel_ascendants(self.descendants(code, parent)) for code, parent in geo_parents.items()}
    for t in [t for t in list_of_tuples['tuples'] if "attr" in t or "obs" in t]:
      aggr_func =  self.tuple_aggregate_function(t, variables)
      if aggr_func is None:
        untouched.append(t)
      else:
        for aggr_key, tt in self.pairs_to_aggregate(t, var_asc, variables):
          if not aggr_key in cumm:
            cumm[aggr_key] = tt
          else:
            var_name = list(t["obs"].keys())[0]
            cumm[aggr_key]["obs"][var_name] = aggr_func([cumm[aggr_key]["obs"][var_name], tt["obs"][var_name]])

    result = list_of_tuples.copy()
    result['tuples'] = untouched + [*cumm.values()]
    
    ret = self._storage_proxy.to_job_cache(job["id"], f"agg", result).get()
    self._pipeline_proxy.aggregate_end(ret, job = job)

  def descendants(self, code, parent):
    codes = self._variables_proxy.get_referential(code).get()
    if codes is None:
      return None
    rel = {t['attr'][code]:t['attrs'][parent] if parent in t['attrs'] else None for t in codes}
    return self.rel_descendants(rel) 

  def rel_descendants(self, rel, desc = {}):
    if len(desc) == 0:
      desc = {c:{c} for c, p in rel.items()}
    added = True
    while added:
      added = False
      for e, dd in desc.items():
        for c, p in rel.items():
          if p in dd and c not in desc[e]:
            desc[e].add(c)
            added = True
    return desc

  def rel_ascendants(self, descendants):
    if descendants is None:
      return None
    asc = {}
    for code, descs in descendants.items():
      for desc in descs:
        if not desc in asc:
          asc[desc] = {code}
        else:
          asc[desc].add(code)
    return asc

  def aggregate_function(self, unit):
    if unit is None:
      return lambda values: None 

    unit = unit.lower().replace(" ", "")
    if unit in ['people', 'number', 'qty', 'days']:
        return np.sum
    elif unit in ['comma_list']:
      return ','.join
    else:
      return lambda values: None 
   

  def tuple_aggregate_function(self, t, variables):
    if ("obs" in t 
      and len(t["obs"].keys()) > 0 
      and list(t["obs"].keys())[0] in variables 
      and "unit" in variables[list(t["obs"].keys())[0]] ):
        return self.aggregate_function(variables[list(t["obs"].keys())[0]]["unit"])
    else:
        return None

  def pairs_to_aggregate(self, t, var_asc, variables):
    if not "attrs" in t or len(t["attrs"].keys()) == 0:
      yield ('', t)
    else:
      keys = [attr for attr in t["attrs"].keys() if variables[attr]["type"] in ["characteristic", "referential", "geo_referential", "date", "referential_parent"]]
      keys.sort()
      # adding identity aggregation 
      yield (json.dumps({list(t["obs"].keys())[0]:[(key,t["attrs"][key]) for key in keys]}, cls=JsonEncoder), t)

      # adding ascendants aggregation
      for code_var, ascendants in var_asc.items():
        if code_var in t["attrs"] and t["attrs"][code_var] in ascendants:
          code =  t["attrs"][code_var]
          for asc in ascendants[code]:
            c = copy.deepcopy(t)
            c["attrs"][code_var] = asc
            keys = [attr for attr in c["attrs"].keys() if variables[attr]["type"] in ["characteristic", "referential", "geo_referential", "date", "referential_parent"]]
            keys.sort()
            if(code != asc): # we have to remove the idenntity since it was already added
              yield (json.dumps({list(t["obs"].keys())[0]:[(key,c["attrs"][key]) for key in keys]}, cls=JsonEncoder), c)

  def prepare_tuples(self, tuples):
    tts = {
      "tuples":[],
      "scope":{"update_scope":{}}
    }
    for p, tt in tuples.items():
      if tt is not None:
        # getting tuples from cache
        tt = tt.value()
        tts['tuples'].extend(tt['tuples'])
        for u in tt['scope']['update_scope']:
          v = u["value"] if type(u["value"]) in [list, set] else [u["value"]]
          if u['variable'] in tts['scope']['update_scope'] :
             tts['scope']['update_scope'][u['variable']].update(v)
          else:
             tts['scope']['update_scope'][u['variable']] = set(v)

    tts['scope']['update_scope'] = [{"variable":k, "value":list(v)} for k, v in tts['scope']['update_scope'].items()]
    return tts



import subprocess
import os
from . import worker


class ScriptExecutor(worker.Worker):
    def __init__(self, name, orchestrator_ref, settings): 
        super().__init__(name = name, orchestrator_ref = orchestrator_ref, settings = settings)
         
    def on_start(self):
        super().on_start()

    def script_path(self, script_type, script_name):
        return self.pandem_path("files", "scripts", script_type, f"{script_name}.{script_type}")

    def execute(self, script_name, script_type, working_dir):
        if not os.path.exists(working_dir):
            os.makedirs(working_dir)
        if script_type == "R":
            print(f'R script will be runned with: {self.script_path(script_type = script_type, script_name = script_name)}')
            subprocess.run(['Rscript', '--vanilla',  self.script_path(script_type = script_type, script_name = script_name)], cwd=working_dir) 
            print('R script has just been run')
        else: 
            raise NotImplementedError("So far only R scripts are supported")



import numpy as np

def df_transform(df):
  df["parent"] = np.where(df["code"].str.len()==2, None, df["code"].str[0:-1])
  df["level"] =  np.where(
    df["code"].str.len()==2, "Country", np.where(
    df["code"].str.len()==3, "NUTS-1",  np.where(
    df["code"].str.len()==4, "NUTS-2", np.where(
    df["code"].str.len()==5, "NUTS-3", "Other"
    ))))
  return df



from abc import ABC, abstractmethod, ABCMeta
import pykka
import os
import time
import threading
from datetime import datetime, timedelta
import pandas as pd
import numpy as np

class Worker(pykka.ThreadingActor):
    '''
    The Worker abstract class implements a Pykka actor and gives a blueprint for actors that are launched by the orchestrator actor.

    Args:
        name (str): The name given by the orchestration actor to this worker.
        orchestrator_ref: The orchestrator reference.
        setting: The package default configuration values.
    '''
    __metaclass__ = ABCMeta  

    

    def __init__ (self, name, orchestrator_ref, settings):
        super().__init__()
        self.name = name
        self.settings = settings
        self._orchestrator_proxy = orchestrator_ref.proxy()
        self._self_proxy = self.actor_ref.proxy()
        self._actions = []
        
    
    def on_start(self):
        '''Extends the Pykka on_start method for optional setup after initialisation and before processing messages.'''
        heartbeat_repeat = Repeat(timedelta(seconds=20))
        self.register_action(heartbeat_repeat, self.send_heartbeat)
        threading.Thread(target=self.actor_loop).start()

    def send_heartbeat(self):
        '''A class method that sends heartbeat meassages to the orchestrator'''
        self._orchestrator_proxy.get_heartbeat(self.name)


    def actor_loop(self):
        while True:
            time.sleep(0.01)
            #print(f'last execution list is: {last_executions}')
            next_executions = [action['repeat'].next_exec() for action in self._actions]
            #print(f'next execution list is: {next_executions}')
            next_deltas = [(next_time - datetime.now()).total_seconds() for next_time in next_executions]
            #print(f'next deltas list is: {next_deltas}')
            next_action_index = next_deltas.index(min(next_deltas)) #return the index of the first min value
            #print(f'next action index is: {next_action_index}')

            next_action = self._actions[next_action_index]
            #print(f'next action is: {next_action}')
            if datetime.now() > next_executions[next_action_index]:
                next_action["repeat"].last_exec = datetime.now()
                next_action["func"]()
                if next_action["oneshot"]:
                  self._actions.pop(next_action_index)
                else:
                  self._actions[next_action_index] = next_action 


    def pandem_path(self, *args):
        '''A class method to get the absolute path from relative paths components'''
        return os.path.join(os.getenv('PANDEM_HOME'), *args)

    def register_action(self, repeat, action, oneshot = False):
        self._actions.append({'repeat': repeat, 'func': action, 'oneshot':oneshot })
        
    def staging_path(self, job_id, *args):
        return self.pandem_path(f'files/staging', str(int(job_id)), *args)


class Repeat:
        def __init__(self, tdelta, start_hour=None, end_hour=None, last_exec=None):
            self.tdelta = tdelta
            self.start_hour = start_hour
            self.end_hour = end_hour
            self.last_exec = last_exec
        
        def next_exec(self): #last_exec:datetime.datetime()
            if pd.isna(self.last_exec):
              self.last_exec = None
            if self.last_exec is None:
              next_exec = datetime.now()
            else :
              next_exec = self.last_exec + self.tdelta

              if self.start_hour is None:
                st = 0
              else:
                st = self.start_hour

              if self.end_hour is None:
                en = 25
              else:
                en = self.end_hour

              # if next execution is outside of the hour window, next execution will be next day at the beginning of the period
              if next_exec.hour < st or next_exec.hour > en:
                next_exec = datetime.combine(next_exec.date(), datetime.min.time()) + timedelta(days = 1, hours = st)
            return next_exec




    
  
